#ifndef GDS_VECTOR_LIBRARY_HH
#define GDS_VECTOR_LIBRARY_HH

//==================================  Figure out the platform
#if defined(__GNUG__) && (defined(__i386__) || defined(__x86_64__))
  #if defined(__AVX__)
//#define GDS_USE_AVX 1
    #define GDS_USE_SSE4 1
  #elif defined(__SSE4_1__)
    #define GDS_USE_SSE4 1
  #elif defined(__SSE3__)
    #define GDS_USE_SSE3 1
  #elif defined(__SSE2__)
    #define GDS_USE_SSE2 1
  #else
    #warning "SSE2 not enabled"
    #define GDS_USE_BASE 1
  #endif
#else 
  #warning "Not a gnu compilation"
  #define GDS_USE_BASE 1
#endif // end test of i386/gcc

//==================================  Useful data types
typedef double vec_result;

#if defined(GDS_USE_SSE2) || defined(GDS_USE_SSE3) || defined(GDS_USE_SSE4)
#define GDS_VEC_LENGTH 16
typedef double vecd128 __attribute__ ((vector_size(GDS_VEC_LENGTH)));
typedef float  vecf128 __attribute__ ((vector_size(GDS_VEC_LENGTH)));
union vd2map {
  vecd128 v;
  double d[2];
};
#endif // end test of sse2

#ifdef GDS_USE_AVX
#define GDS_VEC_LENGTH 32
typedef double vecd256 __attribute__ ((vector_size(GDS_VEC_LENGTH)));
typedef float  vecf256 __attribute__ ((vector_size(GDS_VEC_LENGTH)));
#endif // end test of avx




//======================================  Inner product.
#if defined(GDS_USE_SSE2) || defined(GDS_USE_SSE3) || defined(GDS_USE_SSE4)
inline vec_result
vdot(const double* a, const double* b, size_t n) {
  vec_result sum = 0.0;
#if GDS_USE_SSE2
  vd2map sumu;
#else
  double dvec;
#endif

  //------------------------------------  Check alignment for sse2.
  if (((long(a) ^ long(b)) & (GDS_VEC_LENGTH - 1)) == 0) {

    //----------------------------------  Calculate 1 word if unaligned.
    size_t nLeft = n;
    if ((long(a) & sizeof(double)) != 0) {
      sum = (*a++) * (*b++);
      nLeft--;
    }

    //----------------------------------  Loop over pairs with sse2
    size_t nLoop = nLeft >> 1;
    if (nLoop) {
      const vecd128* av = reinterpret_cast<const vecd128*>(a);
      const vecd128* bv = reinterpret_cast<const vecd128*>(b);
      register vecd128 sumv = *av * *bv;
      for (size_t i=1; i<nLoop; ++i) sumv += av[i] * bv[i];
#ifdef GDS_USE_SSE2
      sumu.v  = sumv;
      sum += sumu.d[0] + sumu.d[1];
#else
      asm("haddpd %1, %1;"
	  "movsd %1, %0"
	  : "=m" (dvec)
	  : "x" (sumv)
	  );
      sum += dvec;
#endif
    }

    //----------------------------------  Add last entry.
    if ((nLeft & 1) != 0) sum += a[nLeft-1] * b[nLeft-1];
  }

  //------------------------------------  Use a single loop if unaligned.
  else {
    size_t nVec = n >> 1;
    if (nVec) {
      const vecd128* av;
      const double* bv;
      if ((long(b) & sizeof(double)) != 0) {
	av = reinterpret_cast<const vecd128*>(a);
	bv = b;
      } else {
	av = reinterpret_cast<const vecd128*>(b);
	bv = a;
      }
      
      register vecd128 sumv = {0.0, 0.0};
      register vecd128 temp;
      for (size_t i=0; i<nVec; i++) {
#ifdef GDS_USE_SSE2
	asm("movups %1, %0": "=x" (temp) : "m" (bv[0]));
#else
	asm("lddqu %1, %0": "=x" (temp) : "m" (bv[0]));
#endif
	bv   += 2;
	temp *= av[i];
	sumv += temp;
      }
#ifdef GDS_USE_SSE2
      sumu.v  = sumv;
      sum += sumu.d[0] + sumu.d[1];
#else
      asm("haddpd %1, %1;"
	  "movsd %1, %0"
	  : "=m" (dvec)
	  : "x" (sumv)
	  );
      sum += dvec;
#endif
    }
    if ((n & 1) != 0) sum += a[n-1] * b[n-1];
  }
  return sum;
}

#else
//======================================  Base dot product function.
inline vec_result
vdot(const double* a, const double* b, size_t n) {
  if (!n) return 0.0;

  vec_result sum = *a * *b;
  for (size_t i=1; i<n; i++) sum += a[i] * b[i];
  return sum;
}
#endif // end vdot version

//======================================  Multiply vectors
#if defined(GDS_USE_SSE2)
inline void
vmul3(const double* a, const double* b, double* c, size_t n) {
  //------------------------------------  Test 3-fold alignment
  if ((((long(a)^long(b)) | (long(a)^long(c))) & (GDS_VEC_LENGTH - 1)) == 0) {

    //----------------------------------  One double multiply if phase wrong
    size_t nLeft = n;
    if ((long(a) & sizeof(double)) != 0) {
      *c++ = *a++ * *b++;
      nLeft--;
    }

    //----------------------------------  Loop over sse2 multiplies
    size_t nLoop = nLeft >> 1;
    if (nLoop) {
      const vecd128* av = reinterpret_cast<const vecd128*>(a);
      const vecd128* bv = reinterpret_cast<const vecd128*>(b);
      vecd128* cv = reinterpret_cast<vecd128*>(c);
      for (size_t i=0; i<nLoop; ++i) cv[i] = av[i] * bv[i];
    }

    //----------------------------------  Last element.
    if ((nLeft & 1) != 0) c[nLeft-1] = a[nLeft-1] * b[nLeft-1];
  }

  //------------------------------------   tight loop...
  else {
    for (size_t i=0; i<n; ++i) c[i] = a[i] * b[i];
  }
}

#endif  // vmul3 version.

//======================================  Multiply vectors
#if defined(GDS_USE_SSE2)
inline void
vmul(const double* a, double* c, size_t n) {
  if (((long(a) ^ long(c)) & (GDS_VEC_LENGTH - 1)) == 0) {
    size_t nLeft = n;
    if ((long(a) & sizeof(double)) != 0) {
      *c++ *= *a++;
      nLeft--;
    }

    size_t nLoop = nLeft >> 1;
    if (nLoop) {
      const vecd128* av = reinterpret_cast<const vecd128*>(a);
      vecd128* cv = reinterpret_cast<vecd128*>(c);
      for (size_t i=0; i<nLoop; ++i) cv[i] *= av[i];
    }

    if ((nLeft & 1) != 0) c[nLeft-1] *= a[nLeft-1];
  }

  else {
    for (size_t i=0; i<n; ++i) c[i] *= a[i];
  }
}

#else
//======================================  Scaler evaluation version.
inline void
vmul(const double* a, double* c, size_t n) {
  for (size_t i=0; i<n; ++i) c[i] *= a[i];
}

#endif  // end vmul version.

//======================================  Scale a vector.
#if defined(GDS_USE_SSE2)
inline void
vscale(double* a, double b, size_t n) {
  size_t nLeft = n;
  if ((long(a) & sizeof(double)) != 0) {
    *a++ *= b;
    nLeft--;
  }

  size_t nLoop = nLeft >> 1;
  if (nLoop) {
    vecd128* av = reinterpret_cast<vecd128*>(a);
    register vecd128  bv = {b, b};
    for (size_t i=0; i<nLoop; ++i) av[i] *= bv;
  }

  if ((nLeft & 1) != 0) a[nLeft-1] *= b;
}

#else

//======================================  Scaler evaluation version.
inline void
vscale(double* a, double b, size_t n) {
  for (size_t i=0; i<n; i++) a[i] *= b;
}
#endif // end vscale version

//====================================  vcmodsq using sse4.1 version.
#if defined(GDS_USE_SSE4)
inline void
vcmodsq(const double* a, double* b, size_t n) {
  if ((long(a) & sizeof(double)) == 0) {
    for (size_t i=0; i<n; ++i) {
      asm("movdqa %1, %%xmm0;"
	  "dppd $51, %%xmm0, %%xmm0;"
	  "movsd %%xmm0, %0"
	  : "=m" (b[i])
	  : "g"  (*a)
	  : "xmm0"
	  );
      a += 2;
    }
  } else {
    for (size_t i=0; i<n; ++i) {
      asm("lddqu %1, %%xmm0;"
	  "dppd $51, %%xmm0, %%xmm0;"
	  "movsd %%xmm0, %0"
	  : "=m" (b[i])
	  : "g"  (*a)
	  : "xmm0"
	  );
      a += 2;
    }
  }
}

#elif defined(GDS_USE_SSE3)
//====================================  vcmodsq using sse2.
inline void
vcmodsq(const double* a, double* b, size_t n) {
  vd2map sumu;
  if ((long(a) & sizeof(double)) == 0) {
    for (size_t i=0; i<n; ++i) {
      asm("movdqa %1, %%xmm0;"
	  "mulpd %%xmm0, %%xmm0;"
	  "haddpd %%xmm0, %%xmm0;"
	  "movsd %%xmm0, %0"
	  : "=m" (b[i])
	  : "g"  (*a)
	  : "xmm0"
	  );
      a += 2;
    }
  } else {
    for (size_t i=0; i<n; ++i) {
      asm("lddqu %1, %%xmm0;"
	  "mulpd %%xmm0, %%xmm0;"
	  "haddpd %%xmm0, %%xmm0;"
	  "movsd %%xmm0, %0"
	  : "=m" (b[i])
	  : "g"  (*a)
	  : "xmm0"
	  );
      a += 2;
    }
  }
}
#else
//======================================  Scaler evaluation version.
inline void
vcmodsq(const double* a, double* b, size_t n) {
  for (size_t i=0; i<n; i++) {
    b[i] = a[0] * a[0] + a[1] * a[1];
    a += 2;
  }
}
#endif // end vcmodsq version


#endif // !defined(gds_vector_library_hh)
